/*
  GPDMA driver for LPC1700 Microcontrollers on FreeRTOS
  Copyright (C) 2010  DF9DQ
  df9dq at ov-lennestadt dot de

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <inttypes.h>

#include <FreeRTOS.h>
#include <task.h>
#include <semphr.h>
#include <queue.h>

#include "LPC17xx.h"
#include "dma.h"


  /* Module variables */
static xSemaphoreHandle g_xDmacSemaphore;   // Semaphore for exclusive access to DMAC resources.


static dma_CallbackFromISR callbacks[DMAC_NUM_CHANNELS];


  /*
   * GPDMA interrupt handler
   */
void GPDMA_IRQHandler( void );
void GPDMA_IRQHandler( void )
{
  uint32_t        regVal;
  uint32_t        n;


    regVal = LPC_GPDMA->DMACIntTCStat;
    LPC_GPDMA->DMACIntTCClr = regVal;

  // Report terminal count?
  for( n = 0 ; n < DMAC_NUM_CHANNELS ; n++ )
  {
    if( regVal & (1u << n) )
    {
      /* Callback handler */
      if (callbacks[n]) {
          callbacks[n](DMA_STATUS_TERMINAL_COUNT, NULL);
      }
    }
  }

    regVal = LPC_GPDMA->DMACIntErrStat;
    if (regVal) {
    // (DMACIntErrClr is a write-only register!)
//TODO    GPDMA->DMACIntErrClr = regVal & DMACIntErrClr_MASK;
    }
}


  /*
   * General DMA initialization
   * Create DMA Semaphores
   * Install DMA Interrupt Handler
   *
   */
void dmaInit( void )
{
  // Create semaphore for exclusive DMAC access
  vSemaphoreCreateBinary( g_xDmacSemaphore );
  xSemaphoreGive( g_xDmacSemaphore );

  // Install interrupt handler
  DMAC_INSTALL_INT_HANDLER();
}



  /*
   * Resume DMAC operation (or first time power-up)
   *
   * start/resume DMA operation
   * If no exclusive access is possible within 10 ms DMAC_BUSY is reported
   * DMAC_ACTIVE (already enabled) or DMAC_OK (just started) otherwise
   */
DMAC_STATUS_CODE dmaResume( void )
{
  DMAC_STATUS_CODE result;


  result = DMAC_BUSY;

  // Access must be exclusive
  if( xSemaphoreTake( g_xDmacSemaphore, ( portTickType ) 10 ) == pdTRUE )
  {
    // Already enabled?
    if (LPC_GPDMA->DMACConfig & 1)
      result = DMAC_ACTIVE;
    else
    {
      // Enable the clock for the DMA block
      DMAC_CLOCK_ENABLE();

      // Enable the GPDMA (little endian mode)
      LPC_GPDMA->DMACConfig = 0x01;

      result = DMAC_OK;
    }

    // Release the semaphore
    xSemaphoreGive( g_xDmacSemaphore );
  }

  return result;
}



  /*
   * Suspend DMAC operation
   */
DMAC_STATUS_CODE dmaSuspend( void )
{
  DMAC_STATUS_CODE result;


  result = DMAC_BUSY;

  // Access must be exclusive
  if( xSemaphoreTake( g_xDmacSemaphore, ( portTickType ) 10 ) == pdTRUE )
  {
    // One of the channels still enabled?
    if (LPC_GPDMA->DMACEnbldChns & DMAC_CHANNEL_MASK)
      result = DMAC_BUSY;
    else
    {
      // Disable the DMA
      LPC_GPDMA->DMACConfig = 0;

      // Disable the DMA clock
      DMAC_CLOCK_DISABLE();

      result = DMAC_OK;
    }

    // Release the semaphore
    xSemaphoreGive( g_xDmacSemaphore );
  }

  return result;
}


/* Hack...  Set callback */
void dmaSetCallback (DMAC_HANDLE handle, dma_CallbackFromISR callback)
{
    uint8_t channel = (uint8_t)handle;

    callbacks[channel] = callback;
}


  /*
   * Initialize a DMA channel for transfer.
   * You must create a linked list in SRAM, then pass a pointer
   * to the first list element.
   */
DMAC_HANDLE dmaStartChannel( uint32_t config, const struct strLLI *lli, dma_CallbackFromISR callback)
{
  int channel;
  DMAC_STATUS_CODE result;


  result = DMAC_BUSY;

  // Access must be exclusive
  if (xSemaphoreTake( g_xDmacSemaphore, ( portTickType ) 5 ) == pdTRUE)
  {
    // Check if DMA is enabled at all
    if( !(LPC_GPDMA->DMACConfig & 1) )
      result = DMAC_SUSPENDED;
    else
    {
      // Find a free channel.
      for( channel = 0 ; channel < DMAC_NUM_CHANNELS ; channel++ )
      {
        if (!(LPC_GPDMAChannel[channel].DMACCConfig & 1)) {
            break;
        }
      }
      if( channel < DMAC_NUM_CHANNELS )
      {
        // Remember the callback
        callbacks[ channel ] = callback;

        // Clear any pending interrupt flags of the channel
        LPC_GPDMA->DMACIntTCClr = 1 << channel;
        LPC_GPDMA->DMACIntErrClr = 1 << channel;

        // Copy information from first LLI directly into DMAC channel registers.
        LPC_GPDMAChannel[channel].DMACCSrcAddr = lli->source;
        LPC_GPDMAChannel[channel].DMACCDestAddr = lli->destination;
        LPC_GPDMAChannel[channel].DMACCLLI = lli->next;
        LPC_GPDMAChannel[channel].DMACCControl = lli->control;

        // Set configuration word (this enables the channel)
        LPC_GPDMAChannel[channel].DMACCConfig = config;

        result = channel;
      }
    }

    // Release the semaphore
    xSemaphoreGive( g_xDmacSemaphore );
  }

  return result;
}


  /**
   * Return a DMAC handle, which was previously obtained by \ref dmaStartChannel.
   */
DMAC_STATUS_CODE dmaReleaseHandle( DMAC_HANDLE handle )
{
  DMAC_STATUS_CODE result = DMAC_OK;


  // It must be a valid handle
  if( (handle < 0) || (handle >= DMAC_NUM_CHANNELS) )
    result = DMAC_INVALID_HANDLE;
  else
  {
    // Disable the channel (ignore data in FIFO)
    LPC_GPDMAChannel[handle].DMACCConfig = 0;
  }

  return result;
}

