/*
  GPDMA driver for LPC1700 Microcontrollers on FreeRTOS
  Copyright (C) 2007  DF9DQ
  df9dq at ov-lennestadt dot de

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __DMA_H
#define __DMA_H


#include <inttypes.h>
#include <stdbool.h>

#include <queue.h>
#include <semphr.h>



  /* Some implementation specific definitions */

  // The number of channels in this implementation.
  // This is '8' for the LPC1700
#define DMAC_NUM_CHANNELS           (8)

  // Mask containing a '1' for every physical channel
#define DMAC_CHANNEL_MASK           (0x000000FF)

  // Commands to enable and disable the clock to the DMAC (power saving)
#define DMAC_CLOCK_ENABLE()         {LPC_SC->PCONP |= (1 << 29);}
#define DMAC_CLOCK_DISABLE()        {LPC_SC->PCONP &= ~(1 << 29);}

  // Install the interrupt handler
#define DMAC_INSTALL_INT_HANDLER()                                    \
    {                                                                 \
      NVIC_SetPriority(DMA_IRQn, configDMA_INTERRUPT_PRIORITY);       \
      NVIC_EnableIRQ( DMA_IRQn );                                     \
    }


  /* Definitions specific to the PrimeCell DMAC block */

  // Fields of DMAC control word
#define DMAC_BURSTSIZE_1            (0u)
#define DMAC_BURSTSIZE_4            (1u)
#define DMAC_BURSTSIZE_8            (2u)
#define DMAC_BURSTSIZE_16           (3u)
#define DMAC_BURSTSIZE_32           (4u)
#define DMAC_BURSTSIZE_64           (5u)
#define DMAC_BURSTSIZE_128          (6u)
#define DMAC_BURSTSIZE_256          (7u)

#define DMAC_TRANSFERWIDTH_8        (0u)
#define DMAC_TRANSFERWIDTH_16       (1u)
#define DMAC_TRANSFERWIDTH_32       (2u)

#define DMAC_NO_INCREMENT           (0u)
#define DMAC_INCREMENT              (1u)

#define DMAC_TC_ON                  (1u)
#define DMAC_TC_OFF                 (0u)


  // Peripheral ID for DMAC configuration
#define DMAC_ID_SSP0_TX             (0u)
#define DMAC_ID_SSP0_RX             (1u)
#define DMAC_ID_SSP1_TX             (2u)
#define DMAC_ID_SSP1_RX             (3u)
#define DMAC_ID_ADC                 (4u)
#define DMAC_ID_I2S_0               (5u)
#define DMAC_ID_I2S_1               (6u)
#define DMAC_ID_DAC                 (7u)
#define DMAC_ID_UART0TX_MAT00       (8u)
#define DMAC_ID_UART0RX_MAT01       (9u)
#define DMAC_ID_UART1TX_MAT10       (10u)
#define DMAC_ID_UART1RX_MAT11       (11u)
#define DMAC_ID_UART2TX_MAT20       (12u)
#define DMAC_ID_UART2RX_MAT21       (13u)
#define DMAC_ID_UART3TX_MAT30       (14u)
#define DMAC_ID_UART3RX_MAT31       (15u)
#define DMAC_ID_DUMMY               (0u)

  // Flow control for DMAC configuration
#define DMAC_FLOW_M2M_DMAC          (0u)
#define DMAC_FLOW_M2P_DMAC          (1u)
#define DMAC_FLOW_P2M_DMAC          (2u)
#define DMAC_FLOW_P2P_DMAC          (3u)
#define DMAC_FLOW_P2P_DESTPER       (4u)
#define DMAC_FLOW_M2P_PER           (5u)
#define DMAC_FLOW_P2M_PER           (6u)
#define DMAC_FLOW_P2P_SRCPER        (7u)

  // Interrupt enable bits for DMAC configuration
#define DMAC_ERROR_INT_ON           (1u)
#define DMAC_ERROR_INT_OFF          (0u)
#define DMAC_TERMINAL_COUNT_INT_OFF (0u)
#define DMAC_TERMINAL_COUNT_INT_ON  (1u)


  // Linked-list item
struct strLLI
{
  uint32_t  source;         // Start of source area
  uint32_t  destination;    // Start of destination area
  uint32_t  next;           // Address of next strLLI in chain
  uint32_t  control;        // DMACCxControl register
};


typedef enum
{
  DMAC_BUSY = -1,           // DMAC busy
  DMAC_SUSPENDED = -2,      // DMAC is inactive
  DMAC_ACTIVE = -3,         // DMAC is already active
  DMAC_INVALID_HANDLE = -4, // DMAC handle was invalid

  DMAC_OK = 0               // DMAC ok

} DMAC_STATUS_CODE;


typedef enum
{
  DMAC_TRANSFER_OK = 0,     // DMAC transfer done without errors

} DMAC_EVENT_CODE;


typedef int32_t DMAC_HANDLE;


typedef struct tagDMAC_EVENT
{
  DMAC_HANDLE       handle;
  DMAC_EVENT_CODE   event;
} DMAC_EVENT;



  /*! Assemble the control word for DMACCxControl.
   */
#define DMAC_CONTROL_WORD(                                                          \
      transfer_size, sbsize, dbsize, swidth, dwidth,                                \
      src_inc, dest_inc, terminal_count_int )                                       \
                (                                                                   \
                transfer_size                                                       \
            |   ((sbsize) << 12)                                                    \
            |   ((dbsize) << 15)                                                    \
            |   ((swidth) << 18)                                                    \
            |   ((dwidth) << 21)                                                    \
            |   ((src_inc) << 26)                                                   \
            |   ((dest_inc) << 27)                                                  \
            |   ((terminal_count_int) << 31)                                        \
                )

  /*! Assemble the configuration word for DMACCxConfiguration.
   */
#define DMAC_CONFIG_WORD( source_id, dest_id, flow, error_int, terminal_count_int ) \
                (                                                                   \
                1                                                                   \
            |   ((source_id) << 1)                                                  \
            |   ((dest_id) << 6)                                                    \
            |   ((flow) << 11)                                                      \
            |   ((error_int) << 14)                                                 \
            |   ((terminal_count_int) << 15)                                        \
                )

/** Status code for the callback function. */
enum {
    DMA_STATUS_TERMINAL_COUNT,
    DMA_STATUS_ERROR,
};

/** DMA callback function. */
typedef void (*dma_CallbackFromISR)(uint8_t status, void *parameter);


  /*! \brief Initialize GPDMA block.
   *
   * Must be called once during system initialization before any other DMAC
   * function or data is used.
   */
void dmaInit( void );

  /*! \brief Resumes DMAC operation.
   *
   * Must be called after fnDmacInit or after fnDmacSuspend, before the DMAc
   * can be used.
   *
   * \return Usually returns DMAC_OK to indicate success.
   *         If already enabled, it returns DMAC_ACTIVE.
   */
DMAC_STATUS_CODE dmaResume( void );

  /*! \brief Suspends DMAC operation.
   *
   * May be called whenever the DMAC block is no longer needed.
   * Puts the block in a low power state.
   *
   * \return Usually returns DMAC_OK to indicate success.
   *         If you try to disable the DMAC while there are still enabled
   *         channels, the function returns DMAC_BUSY.
   */
DMAC_STATUS_CODE dmaSuspend( void );

  /*! \brief Initialize and start a DMAC channel.
   *
   * You pass parameters to describe the DMA task to be done.
   * It is possible to get a message once a selected DMA event occurs,
   * like error or terminal count.
   *
   * An alternative method to detect the end of a DMA operation is to poll the
   * channel status after you started it.
   *
   * \param config Configuration word for DMACCxConfiguration. Can be created easily
   *        with DMAC_CONFIG_WORD.
   * \param lli Pointer to the first linked-list item (LLI). It can be a single item,
   *        or it can be the first LLI of a longer list.
   * \param semaphore_handle This semaphore will be 'given' by the DMA handler
   *        once the transfer is over.
   * \return >=0: The channel number allocated for the transfer.
   *         <0 : An error number (DMAC_STATUS_CODE)
   */
DMAC_HANDLE dmaStartChannel( uint32_t config, const struct strLLI *lli, dma_CallbackFromISR callback);

  /*! \brief Explicitely return a DMAC channel handle.
   *
   * Return a DMAC handle, which was previously obtained by \ref dmaStartChannel.
   *
   * \param handle DMAC channel handle.
   *
   * \return Returns DMAC_OK to indicate success.
   *         Returns DMAC_INVALID_HANDLE to indicate an error.
   */
DMAC_STATUS_CODE dmaReleaseHandle( DMAC_HANDLE handle );

/* Hack... */
void dmaSetCallback (DMAC_HANDLE handle, dma_CallbackFromISR callback);

#endif  // #ifndef __DMA_H

