/* Copyright (c) 2010, DF9DQ
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list of conditions
 * and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions
 * and the following disclaimer in the documentation and/or other materials provided with the
 * distribution.
 * Neither the name of the author nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



#include <stdint.h>
#include <stdbool.h>

#include "FreeRTOS.h"
#include "queue.h"
#include "semphr.h"
#include "timers.h"
#include "task.h"

#include "LPC17xx.h"
#include "init.h"
#include "i2c.h"
#include "si570.h"
#include "params.h"
#include "lpcusb/usbapi.h"
#include "../params.h"


    /* Default bus address */
#define SI570_ADDRESS         (0x55 << 1)

    /* Message queue */
#define SI570_QUEUE_LENGTH    (5)
xQueueHandle si570_queue;


typedef struct __attribute__((packed)) {
    uint8_t opcode;
    union {
        uint32_t freq1121;
        uint8_t regs[6];
    };
} si570_Message;


/** Si570 queue messages */
enum {
    SI570_MESSAGE_SET_FREQUENCY,
    SI570_MESSAGE_GET_VIRTUAL_REGISTERS,
    SI570_MESSAGE_GET_REAL_REGISTERS,
    SI570_MESSAGE_GET_FACTORY_STARTUP_REGISTERS,
    SI570_MESSAGE_SET_VIRTUAL_REGISTERS,
    SI570_MESSAGE_SET_REAL_REGISTERS,
    SI570_MESSAGE_NEW_PARAMS,                           /**< Parameters have changed. */
    SI570_MESSAGE_SAVE_PARAMS,                          /**< Update parameters in flash */
};



    /* I2C bus address. Automatically detected by firmware. */
static uint8_t si570_address;
    /* Current *virtual* frequency (11.21). This is what the host sees! */
static uint32_t si570_currentFrequency1121;
    /* Si570 anchor frequency. Real frequency of Si570. May differ from
     * the frequency of the host by the automatically selected prescaler.
     */
static uint32_t si570_anchorFrequency;
    /* Value of HS_DIV at anchor frequency */
static uint8_t si570_anchorHsDiv;
    /* Value of N1 at anchor frequency */
static uint8_t si570_anchorN1;
    /* Current prescaler setting (1, 4, 16, 64) */
static uint32_t si570_current_prescaler;
    /* Current content of real Si570 registers 7...12 */
static uint8_t si570_regs[6];
    /* Current content of virtual Si570 registers 7...12 */
static uint8_t si570_vregs[6];
    /* Content of Si570 registers after power-up (factory defaults) */
static uint8_t si570_factory_default_regs[6];
    /* Flag that indicates register reset condition (virtual registers) */
static _Bool si570_initialized;
    /* One-shot timer to trigger flash parameter update */
static xTimerHandle si570_paramTimer;


/* Forward declarations */

static bool si570_findHsdivN1 (uint32_t freq, uint32_t *hsdiv, uint32_t *n1, bool check_limits);



/** Set the virtual Si570 registers.
 *
 *  \param [in] regs Set of six register values.
 */
void si570_setVirtualRegisters (uint8_t regs[6])
{
    si570_Message message;

    message.opcode = SI570_MESSAGE_SET_VIRTUAL_REGISTERS;
    message.regs[0] = regs[0];
    message.regs[1] = regs[1];
    message.regs[2] = regs[2];
    message.regs[3] = regs[3];
    message.regs[4] = regs[4];
    message.regs[5] = regs[5];
    xQueueSendToBack(si570_queue, &message, 0);
}



/** Set the value of a single virtual register.
 *
 *  \param[in] regNum Register number
 *  \param[in] value New value for register
 */
void si570_setSingleVirtualRegister (uint8_t regNum, uint8_t value)
{
    if ((regNum == 135) && (value & 1)) {
        /* RECALL bit. Reload factory setup. */
        si570_initialized = false;
    }
}



/** Get the virtual Si570 registers.
 *
 *  The virtual registers are returned in the USB buffer, which will be sent directly via
 *  the (control) endpoint.
 */
void si570_getVirtualRegisters (void)
{
    si570_Message message;

    message.opcode = SI570_MESSAGE_GET_VIRTUAL_REGISTERS;
    xQueueSendToBack(si570_queue, &message, 0);
}



/** Get the factory startup values of the Si570 registers.
 *
 *  The registers are returned in the USB buffer, which will be sent directly via
 *  the (control) endpoint.
 */
void si570_getFactoryDefaultRegisters (void)
{
    si570_Message message;

    message.opcode = SI570_MESSAGE_GET_FACTORY_STARTUP_REGISTERS;
    xQueueSendToBack(si570_queue, &message, 0);
}



/** Get the real Si570 registers.
 *
 *  The real registers are returned in the USB buffer, which will be sent directly via
 *  the (control) endpoint.
 */
void si570_getRegisters (void)
{
    si570_Message message;

    message.opcode = SI570_MESSAGE_GET_REAL_REGISTERS;
    xQueueSendToBack(si570_queue, &message, 0);
}



/** Set the Si570 frequency by value.
 *
 *  Frequency of VCO will be requested frequency minus offset (parameter)
 *
 *  \param [in] frequency1121 Frequency in MHz (11.21 format).
 */
void si570_setFrequency (uint32_t frequency1121)
{
    si570_Message message;

    message.opcode = SI570_MESSAGE_SET_FREQUENCY;
    message.freq1121 = frequency1121 - g_params.freqsubtract1121;
    xQueueSendToBack(si570_queue, &message, 0);
}



/** Save device parameter in flash. */
void si570_saveParams (void)
{
    si570_Message message;

    message.opcode = SI570_MESSAGE_NEW_PARAMS;
    xQueueSendToBack(si570_queue, &message, 0);
}


/*************************************************************************************/


/** Read a single register from Si570.
 */
static int si570_readReg (uint8_t reg, uint8_t *value)
{
    *value = 0;

    if (i2cWriteBuffer (si570_address, (uint8_t *) &reg, sizeof (uint8_t)))
        return -1;

    return i2cReadBuffer (si570_address, (uint8_t *) value, sizeof (uint8_t));
}

/** Write to a single Si570 register.
 */
static int si570_writeReg (uint8_t reg, int value)
{
    uint8_t buffer [2];

    buffer [0] = reg;
    buffer [1] = value;

    return i2cWriteBuffer (si570_address, buffer, sizeof (buffer));
}


/** Write to a consecutive set of Si570 registers. */
static int si570_writeBuffer (uint8_t *buffer, int length)
{
    return i2cWriteBuffer (si570_address, buffer, length);
}


/** Set Si570 I2C address (shifted left by 1 bit) */
static void si570_setAddress (uint8_t address)
{
    si570_address = address;
}

/** Read all registers from Si570.
 *
 *  Registers are stored as both a real copy (si570_regs) and a virtual copy (si570_vregs).
 *  The virtual registers allow the host software to treat the complete frequency range of then
 *  FiFi-SDR as one contiguous range, without the need to care about the prescaler.
 *  Frequencies below 4.85GHz/128/11 = 3.445 MHz are not supported by using si570_vregs.
 *  (i.e. RX frequency = 861.25 kHz)
 */
static void si570_readRegisters (void)
{
    uint32_t hs_div;
    uint32_t n1;
    uint64_t rfreq;
    uint32_t freq;


    /* Read the physical registers. */
    si570_readReg( 7, &si570_regs[0]);
    si570_readReg( 8, &si570_regs[1]);
    si570_readReg( 9, &si570_regs[2]);
    si570_readReg(10, &si570_regs[3]);
    si570_readReg(11, &si570_regs[4]);
    si570_readReg(12, &si570_regs[5]);

    /* Virtual device registers are calculated from the current virtual frequency.
     * Check for the lower and upper limits.
     */
    freq = (si570_currentFrequency1121 / 4) * g_params.virtual_vco_factor;
    if (freq < _11_21(4850.0/(11.0*128.0))) {
        freq = _11_21(4850.0/(11.0*128.0));
    }
    if (freq > _11_21(5670.0/(4.0*1.0))) {
        freq = _11_21(5670.0/(4.0*1.0));
    }
    if (si570_findHsdivN1(freq, &hs_div, &n1, false)) {
        /* Calculate multiplier for virtual registers */
        rfreq = freq;
        rfreq <<= 23;
        rfreq *= hs_div * n1;
        rfreq /= (g_params.freqxtal >> 8);  //TODO Auflösung!

        /* Build the virtual register set */
        si570_vregs[0] = (((hs_div-4) & 7) << 5) | (((n1-1) >> 2) & 0x1F);  /* Reg 7 */
        si570_vregs[1] = (((n1-1) & 0x03) << 6) | ((rfreq >> 32) & 0x3F);   /* Reg 8 */
        si570_vregs[2] = ((rfreq >> 24) & 0xFF);                            /* Reg 9 */
        si570_vregs[3] = ((rfreq >> 16) & 0xFF);                            /* Reg 10 */
        si570_vregs[4] = ((rfreq >> 8) & 0xFF);                             /* Reg 11 */
        si570_vregs[5] = (rfreq & 0xFF);                                    /* Reg 12 */
    }
}


/** Get Si570 I2C address (shifted left by 1 bit). */
uint8_t si570_getAddress (void)
{
    return si570_address;
}


/** Auto-detect the Si570 I2C address. */
static void si570_detectI2cAddress (void)
{
    int i;
    uint8_t buffer[2];


    /* Set a default address in case we do not find the device */
    si570_setAddress(SI570_ADDRESS);

    /* All valid I2C addresses */
    for (i = 0; i < 0x78; i++) {
        buffer[0] = 137;    /* Command: unfreeze VCO */
        buffer[1] = 0;
        if (i2cWriteBuffer ((i << 1), buffer, sizeof(buffer)) == 0) {
            si570_setAddress(i << 1);
            return;
        }
    }
}


/** Set the preselector.
 *
 *  Based on the frequency, one of the four possible preselector outputs
 *  is activated.
 *
 *  \param[in] freq Frequency (format 11.21)
 */
static void si570_setPreselector (uint32_t freq_11_21)
{
    uint32_t freq_11_5;
    int i;
    uint8_t pattern;
    uint32_t search_range;


    /* Convert to 11.5 format */
    freq_11_5 = (freq_11_21 >> 16);

#if (BOARD_TYPE == BOARD_ROLF1)
    if (g_params.presel_mode == 0) {
        /* Select one of the prescaler outputs */
        if (freq_11_5 < g_params.freqrxtrip[0]) {
        }
        else if (freq_11_5 < g_params.freqrxtrip[1]) {
        }
        else if (freq_11_5 < g_params.freqrxtrip[2]) {
        }
        else {
        }
    }

    if ((g_params.presel_mode == 1) || (g_params.presel_mode == 2)) {
        pattern = 0;
        search_range = 0xFFFFFFFF;

        /* Search for the smallest frequency range that contains the current frequency */
        for (i = 0; i < 16; i++) {
            if ((freq_11_21 >= g_params.presel_freq[i][0]) &&
                (freq_11_21 <= g_params.presel_freq[i][1])) {
                if (g_params.presel_freq[i][1] - g_params.presel_freq[i][0] < search_range) {
                    /* Smallest fit so far */
                    search_range = g_params.presel_freq[i][1] - g_params.presel_freq[i][0];
                    pattern = g_params.presel_pattern[i];
                }
            }
        }

        /* Set outputs */
//        GPIO2->FIOCLR = ((~pattern) & 0x0F) << 5;
//        GPIO2->FIOSET = (( pattern) & 0x0F) << 5;
    }

    if (g_params.presel_mode == 2) {
        //TODO  Frequenz per UART senden
    }
#endif

#if (BOARD_TYPE == BOARD_KAI1) || (BOARD_TYPE == BOARD_KAI2)
    if (g_params.presel_mode == 0) {
        /* Select one of the prescaler outputs */
        if (freq_11_5 < g_params.freqrxtrip[0]) {
            LPC_GPIO2->FIOCLR = (1 << 5) | (1 << 6) | (1 << 7) | (1 << 8);
            LPC_GPIO2->FIOSET = (1 << 6);
        }
        else if (freq_11_5 < g_params.freqrxtrip[1]) {
            LPC_GPIO2->FIOCLR = (1 << 5) | (1 << 6) | (1 << 7) | (1 << 8);
            LPC_GPIO2->FIOSET = (1 << 7);
        }
        else if (freq_11_5 < g_params.freqrxtrip[2]) {
            LPC_GPIO2->FIOCLR = (1 << 5) | (1 << 6) | (1 << 7) | (1 << 8);
            LPC_GPIO2->FIOSET = (1 << 5);
        }
        else {
            LPC_GPIO2->FIOCLR = (1 << 5) | (1 << 6) | (1 << 7) | (1 << 8);
            LPC_GPIO2->FIOSET = (1 << 8);
        }
    }

    if ((g_params.presel_mode == 1) || (g_params.presel_mode == 2) || (g_params.presel_mode == 3)) {
        pattern = 0;
        search_range = 0xFFFFFFFF;

        /* Search for the smallest frequency range that contains the current frequency */
        for (i = 0; i < 16; i++) {
            if ((freq_11_21 >= g_params.presel_freq[i][0]) &&
                (freq_11_21 <= g_params.presel_freq[i][1])) {
                if (g_params.presel_freq[i][1] - g_params.presel_freq[i][0] < search_range) {
                    /* Smallest fit so far */
                    search_range = g_params.presel_freq[i][1] - g_params.presel_freq[i][0];
                    pattern = g_params.presel_pattern[i];
                }
            }
        }

        /* Set outputs */
        if (g_params.presel_mode == 3) {
            LPC_GPIO2->FIOCLR = ((~pattern) & 0x07) << 5;
            LPC_GPIO2->FIOSET = (( pattern) & 0x07) << 5;
        }
        else {
            LPC_GPIO2->FIOCLR = ((~pattern) & 0x0F) << 5;
            LPC_GPIO2->FIOSET = (( pattern) & 0x0F) << 5;
        }
    }

    if (g_params.presel_mode == 2) {
        //TODO  Frequenz per UART senden
    }
#endif
}



/** Set the required prescaler (CPLD).
 *
 *  Based on the frequency, one of the four possible prescalers
 *  (1, 4, 16, 64) is selected.
 *
 *  \param[in] freq Frequency (format 11.21)
 *  \param[in,out] prescaler If NULL, the prescaler in CPLD is changed.
 *                           If !=NULL, the prescaler value is returned here, and
 *                           the CPLD is not updated.
 */
static void si570_setPrescaler (uint32_t freq, uint32_t *value)
{
    uint32_t prescaler;
    bool swapIQ;


    /* Determine what prescaler is needed now */
    if (freq < _11_21(4 * 0.2)) {
        prescaler = 64;
    }
    else if (freq < _11_21(4 * 0.8)) {
        prescaler = 16;
    }
    else if (freq < _11_21(4 * 2.5)) {
        prescaler = 4;
    }
    else {
        prescaler = 1;
    }

    if (value == NULL) {
        /* I/Q swapped for 3rd harmonic */
        swapIQ = false;
        if ((freq >= g_params.freq_3rd_harmonic) &&
            (freq < g_params.freq_5th_harmonic)) {
            swapIQ = true;
        }

#if BOARD_TYPE == BOARD_ROLF1
        /* Select prescaler in CPLD */
        switch (prescaler) {
            case 1:     LPC_GPIO2->FIOCLR = (1 << 4) | (1 << 5);
                        break;
            case 4:     LPC_GPIO2->FIOCLR = (1 << 5);
                        LPC_GPIO2->FIOSET = (1 << 4);
                        break;
            case 16:    LPC_GPIO2->FIOCLR = (1 << 4);
                        LPC_GPIO2->FIOSET = (1 << 5);
                        break;
            case 64:    LPC_GPIO2->FIOSET = (1 << 4) | (1 << 5);
                        break;
            default:    break;
        }
#endif

#if (BOARD_TYPE == BOARD_KAI1) || (BOARD_TYPE == BOARD_KAI2)
        /* I/Q swap */
        if (swapIQ) {
            LPC_GPIO1->FIOSET = (1 << 4);
        }
        else {
            LPC_GPIO1->FIOCLR = (1 << 4);
        }

        /* Select prescaler in CPLD */
        switch (prescaler) {
            case 1:     LPC_GPIO1->FIOCLR = (1 << 8) | (1 << 9);
                        break;
            case 4:     LPC_GPIO1->FIOCLR = (1 << 8);
                        LPC_GPIO1->FIOSET = (1 << 9);
                        break;
            case 16:    LPC_GPIO1->FIOCLR = (1 << 9);
                        LPC_GPIO1->FIOSET = (1 << 8);
                        break;
            case 64:    LPC_GPIO1->FIOSET = (1 << 8) | (1 << 9);
                        break;
            default:    break;
        }
#endif
    }
    else {
        /* Return prescaler, but do not touch CPLD. */
        *value = prescaler;
    }
}



/** Return the optimum values for HS_DIV and N1.
 *
 *  For power reasons it is desirable to have the lowest possible DCO
 *  frequency within its operating range of 4850...5670 MHz. For the same
 *  reason, HS_DIV should be as large as possible.
 *
 *  \param[in] freq Frequency (format 11.21)
 *  \param[out] hsdiv Pointer to HS_DIV value (4, 5, 6, 7, 9, 11)
 *  \param[out] n1 Pointer to N1 value (1...128)
 *  \param[in] check_limits Check limitations of real Si570 device
 *
 *  \return true if ok, false if freq not allowed
 */
static bool si570_findHsdivN1 (uint32_t freq, uint32_t *hsdiv, uint32_t *n1, bool check_limits)
{
    const uint8_t hsdiv_val[6] = {4, 5, 6, 7, 9, 11};
    int i, n;


#define DCO_MIN _43_21(4850.0)
#define DCO_MAX _43_21(5670.0)

    /* Check limits of real device. Ignore for virtual device. */
    if (check_limits) {
        /* Handle the special frequency bands first.
        * 970 MHz <= freq <= 1134 MHz       --> HS_DIV = 5, N1 = 1
        * 1213 MHz <= freq <= 1417.5 MHz    --> HS_DIV = 4, N1 = 1
        */
        if ((_11_21(970.0) <= freq) && (freq <= _11_21(1134.0))) {
            *hsdiv = 5;
            *n1 = 1;
            return true;
        }

        if ((_11_21(1213.0) <= freq) && (freq <= _11_21(1417.5))) {
            *hsdiv = 5;
            *n1 = 1;
            return true;
        }

        /* Check valid frequency range */
        if ((freq < _11_21(10.0)) || (freq > SI570_MAX_FREQUENCY))
            return false;
    }

    /* Check with all possible N1 dividers: 1 or even number 2...128 */
    for (n = 1; n <= 128 ; n++) {
        if( (n % 2) && (n > 1) )
            continue;

        /* Check with all six possible HS_DIV values. Start with the large values,
        * and stop at the first match.
        */
        for (i = 0; i <= 5; i++) {
            if (((n * hsdiv_val[i]) * (uint64_t)freq >= DCO_MIN) &&
                ((n * hsdiv_val[i]) * (uint64_t)freq <= DCO_MAX)) {
                *hsdiv = hsdiv_val[i];
                *n1 = n;
                return true;
            }
        }
    }

    /* We should never come here... */
    return false;
}



/** Set a new frequency in Si570.
 *
 *  Calculates the required register settings for the Si570.
 *  Automatically selects a suitable prescaler in the CPLD.
 *
 *  \param[in] freq Frequency (format 11.21)
 */
static void si570_setToNewFrequency (uint32_t frequency)
{
    uint32_t hs_div;
    uint32_t n1;
    uint64_t rfreq;
    bool smooth;
    int32_t delta_f;
    uint8_t txbuf[7];
    uint32_t freq;
    uint32_t prescaler;
    uint8_t regValue;


    si570_initialized = true;

    /* Find the required prescaler (no change yet) */
    si570_setPrescaler(frequency, &prescaler);
    freq = frequency * prescaler;

    /* VHF frequencies: Use 3rd/5th harmonic */
    if (frequency >= g_params.freq_5th_harmonic) {
        freq = freq / 5;
    }
    else if (frequency >= g_params.freq_3rd_harmonic) {
        freq = freq / 3;
    }

    /* Find optimum post dividers */
    if (!si570_findHsdivN1(freq, &hs_div, &n1, true))
        return;

    /* Preselector outputs */
    si570_setPreselector(frequency);

    si570_currentFrequency1121 = frequency;

    /* RFREQ */
    rfreq = freq;
    rfreq <<= 23;
    rfreq *= hs_div * n1;
    rfreq /= (g_params.freqxtal >> 8);  //TODO Auflösung!

    /* Prepare message for I2C bus */
    txbuf[0] = 7;
    txbuf[1] = (((hs_div-4) & 0x07) << 5) | (((n1-1) >> 2) & 0x1F); // Reg 7
    txbuf[2] = (((n1-1) & 0x03) << 6) | ((rfreq >> 32) & 0x3F);     // Reg 8
    txbuf[3] = ((rfreq >> 24) & 0xFF);                              // Reg 9
    txbuf[4] = ((rfreq >> 16) & 0xFF);                              // Reg 10
    txbuf[5] = ((rfreq >> 8) & 0xFF);                               // Reg 11
    txbuf[6] = (rfreq & 0xFF);                                      // Reg 12

    /* Check if we can use "smooth tuning" */
    delta_f = (int32_t)freq - (int32_t)si570_anchorFrequency;
    if( delta_f < 0 )
        delta_f = -delta_f;
    /* Note: 'ppm' requires to divide by 1000000. It's easier to divide
     *       by 1048576 (shift 20 bits) though. This reduces the range for
     *       smooth tuning by 5%.
     */
    smooth =
        delta_f <= ((g_params.smoothtune * (int64_t)si570_anchorFrequency) >> 20);
    /* Smooth tuning is only possible if HS_DIV and N1 do not change. */
    smooth = smooth && (hs_div == si570_anchorHsDiv) && (n1 == si570_anchorN1);

    /* Smooth tuning requires to update the RFREQ registers only.
     * A larger frequency step requires to freeze the oscillator first. In this
     * case the frequency becomes the new anchor.
     */
    if (smooth) {
        txbuf[1] = 8;
        si570_readReg(135, &regValue);
        regValue |= 0x20;
        si570_writeReg(135, regValue);     /* Freeze M = 1 */
        si570_writeBuffer(&txbuf[1], 6);
        si570_readReg(135, &regValue);
        regValue &= ~0x20;
        si570_writeReg(135, regValue);     /* Freeze M = 0 */
    }
    else {
        si570_readReg(137, &regValue);
        regValue |= 0x10;
        si570_writeReg(137, regValue);     /* Freeze */
        si570_writeBuffer(&txbuf[0], 7);
        si570_readReg(137, &regValue);
        regValue &= ~0x10;
        si570_writeReg(137, regValue);     /* Unfreeze */
        si570_writeReg(135, 0x40);         /* NewFreq */

        si570_anchorFrequency = freq;
        si570_anchorHsDiv = hs_div;
        si570_anchorN1 = n1;
    }

    /* Set the prescaler */
    si570_setPrescaler(frequency, NULL);

    /* Read back current registers */
    si570_readRegisters();
}


/** Return current frequency.
 *
 *  Corrected by the offset in use.
 */
uint32_t si570_getFrequency (void)
{
    return si570_currentFrequency1121 + g_params.freqsubtract1121;
}


/** Change frequency through the virtual registers.
 *
 *  Calculate the desired real frequency from the new virtual register set. Then update the system
 *  using the normal way of 'update by frequency'.
 *
 *  TODO: Durch Rundungsfehler werden nicht exakt die gleichen Registerwerte zurückgelesen!
 *
 *  \param[in] regs New values for virtual register set.
 */
static void si570_setVregs (uint8_t regs[6])
{
    uint32_t hs_div;
    uint32_t n1;
    uint64_t rfreq;
    double freq;


    /* Store the new register values */
    si570_vregs[0] = regs[0];
    si570_vregs[1] = regs[1];
    si570_vregs[2] = regs[2];
    si570_vregs[3] = regs[3];
    si570_vregs[4] = regs[4];
    si570_vregs[5] = regs[5];

    /* Extract the usual suspects: HS_DIV, N1, RFREQ */
    hs_div = ((regs[0] >> 5) & 7) + 4;
    n1 = (((regs[0] & 0x1F) << 2) | ((regs[1] >> 6) & 3)) + 1;
    rfreq = (regs[1] & 0x3F);
    rfreq = (rfreq << 32) |
            ((uint32_t)regs[2] << 24) |
            ((uint32_t)regs[3] << 16) |
            ((uint32_t)regs[4] << 8) |
            ((uint32_t)regs[5] << 0);

    /* Calculate the desired RX frequency */
    freq = (double)rfreq / (double)(1UL << 28);
    freq *= (double)g_params.freqxtal / (double)(1UL << 24);
    freq /= (hs_div * n1 * g_params.virtual_vco_factor);

    /* Real frequency (partly virtual actually...) is 4x the RX frequency. */
    si570_setToNewFrequency(_11_21(4.0 * freq));
}


static void si570_OStimerCallback (xTimerHandle timer)
{
    (void) timer;
    si570_Message message;

    message.opcode = SI570_MESSAGE_SAVE_PARAMS;
    xQueueSendToBack(si570_queue, &message, 0);
}


void si570_task (void *pvParameters)
{
    (void) pvParameters;
    si570_Message message;


    si570_currentFrequency1121 = 0;	/* This forces a full register load on first frequency set */
    si570_current_prescaler = 1;
    si570_anchorFrequency = 0;
    si570_anchorHsDiv = 0;
    si570_anchorN1 = 0;

    /* Initialize I2C */
    i2cInit();

    /* Auto-detect the Si570 I2C address */
    si570_detectI2cAddress();

    /* Read factory defaults */
    si570_readRegisters();
    si570_factory_default_regs[0] = si570_regs[0];
    si570_factory_default_regs[1] = si570_regs[1];
    si570_factory_default_regs[2] = si570_regs[2];
    si570_factory_default_regs[3] = si570_regs[3];
    si570_factory_default_regs[4] = si570_regs[4];
    si570_factory_default_regs[5] = si570_regs[5];

    message.opcode = SI570_MESSAGE_SET_FREQUENCY;
    message.freq1121 = g_params.freqstartup;
    xQueueSendToBack(si570_queue, &message, 0);

    /* Prepare one-shot timer for parameter update */
    si570_paramTimer = xTimerCreate((const signed char *)"",
                                    1000/portTICK_RATE_MS,      /* Update flash after 1s */
                                    pdFALSE,                    /* One-shot */
                                    NULL,                       /* No ID */
                                    si570_OStimerCallback);

    while (1) {
        /* Check for new message */
        if (xQueueReceive(si570_queue, &message, portMAX_DELAY)) {
            /* Check desired operation */
            switch (message.opcode) {
            case SI570_MESSAGE_SET_FREQUENCY:
                si570_setToNewFrequency(message.freq1121);
                break;

            case SI570_MESSAGE_GET_VIRTUAL_REGISTERS:
                /* Read Si570 registers, but return virtual registers only! */
                if (si570_initialized) {
                    si570_readRegisters();
                    usbcontrol_send_ep0(6, &si570_vregs[0]);
                }
                else {
                    usbcontrol_send_ep0(6, &si570_factory_default_regs[0]);
                }
                break;

            case SI570_MESSAGE_GET_REAL_REGISTERS:
                /* Read Si570 registers, and return them to host (FiFi-SDR specific) */
                si570_readRegisters();
                usbcontrol_send_ep0(6, &si570_regs[0]);
                break;

            case SI570_MESSAGE_GET_FACTORY_STARTUP_REGISTERS:
                /* Send the factory default registers */
                usbcontrol_send_ep0(6, &si570_factory_default_regs[0]);
                break;

            case SI570_MESSAGE_SET_VIRTUAL_REGISTERS:
                /* Update the virtual registers, and adjust the real frequency accordingly. */
                si570_setVregs(message.regs);
                break;

            case SI570_MESSAGE_SET_REAL_REGISTERS:
                /* Not supported */
                break;

            case SI570_MESSAGE_NEW_PARAMS:
                /* New parameters. Trigger delay after which parameters will be saved. */
                xTimerReset(si570_paramTimer, 0);
                break;

            case SI570_MESSAGE_SAVE_PARAMS:
                /* Store parameters in flash */
                paramsWrite();
                break;
            }
        }
    }
}


/** si570 module init function.
 *
 *  Must be called before the scheduler starts. Initializes the message queue.
 */
void si570_initModule (void)
{
    /* Queue for task communication */
    si570_queue = xQueueCreate((unsigned portBASE_TYPE) SI570_QUEUE_LENGTH,
                               (unsigned portBASE_TYPE) sizeof (si570_Message));
}

