/*
	LPCUSB, an USB device driver for LPC microcontrollers	
	Copyright (C) 2006 Bertrik Sikken (bertrik@sikken.nl)

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	1. Redistributions of source code must retain the above copyright
	   notice, this list of conditions and the following disclaimer.
	2. Redistributions in binary form must reproduce the above copyright
	   notice, this list of conditions and the following disclaimer in the
	   documentation and/or other materials provided with the distribution.
	3. The name of the author may not be used to endorse or promote products
	   derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "usbapi.h"
#include "usbdebug.h"
#include "usbstruct.h"
#include "usbclass.h"
#include "usbhw_lpc.h"
#include "gpdma/dma.h"
#include "i2c/si570.h"
#include "params.h"
#include "softrock.h"
#include "init.h"

#include "LPC17xx.h"

#define CONTROL_SELECTOR(unitid,cs,length) \
        ((unitId == (unitid)) && (controlSelector == (cs)) && (pSetup->wLength == (length)))

#define dB (0x0100)

        
#define usbMAX_SEND_BLOCK		( 20 / portTICK_RATE_MS )
#define usbBUFFER_LEN			( 20 )

#define INCREMENT_ECHO_BY 1
#define BAUD_RATE	115200

#define AUDIO_IN_EP			0x83
#define AUDIO_SYNCH_OUT_EP	0x04

#define MAX_PACKET_SIZE	64

#define LE_WORD(x)		((x)&0xFF),((x)>>8)

// CDC definitions
#define CS_INTERFACE			0x24
#define CS_ENDPOINT				0x25


xQueueHandle usb_queue;

static unsigned char abClassReqData[8];

static struct {
    uint32_t timerModulus;
    uint32_t lastDmaTimestamp;
    uint32_t readIndex;
    int16_t currentVolume;
    uint32_t current_sample_rate;
    uint8_t i2s_active;
} fifiaudio;


/*
 * USB DMA
 */
struct dma_descriptor_type
{
    struct dma_descriptor_type *next;
    uint32_t value1;
    uint32_t *dma_buffer_start;     /* Start address of DMA buffer */
    uint32_t value3;
    uint32_t *isoc_packetsize_memory_address;
};

//TODO  guarantee alignment of 128 bytes!
static struct dma_descriptor_type *usb_udca[32] __attribute__ ((section(".ram2bss"))) __attribute__ ((aligned (128)));

static struct dma_descriptor_type usb_dma_desc[2] __attribute__ ((section(".ram2bss")));
static uint32_t usb_isoc_packets[2] __attribute__ ((section(".ram2bss")));


#define I2S_NUM_SAMPLES (1024)

/** Ring buffer for audio frames. */
static uint32_t audioBuffer[I2S_NUM_SAMPLES + 97] __attribute__((section(".ram2bss")));


const struct strLLI audioLli __attribute__ ((section(".ram2data"))) = {
    .source         = (uint32_t)&LPC_I2S->I2SRXFIFO,
    .destination    = (uint32_t)&audioBuffer[0],
    .next           = (uint32_t)&audioLli,
    .control        = DMAC_CONTROL_WORD(
                            I2S_NUM_SAMPLES,
                            DMAC_BURSTSIZE_4,
                            DMAC_BURSTSIZE_4,
                            DMAC_TRANSFERWIDTH_32,
                            DMAC_TRANSFERWIDTH_32,
                            DMAC_NO_INCREMENT,
                            DMAC_INCREMENT,
                            DMAC_TC_ON),
};


/** Callback for I2S DMA. */
static void fifiaudio_dmaCallbackFromISR (uint8_t status __attribute__((unused)),
                                          void *parameter __attribute__((unused)))
{
    /* Take a time stamp */
    fifiaudio.lastDmaTimestamp = LPC_TIM3->TC;
}


static void i2sStop( void )
{
	// cleanly stop I2S DMA
//    GPDMAChannel[0].DMACCConfig |= (1 << 18);
//    while( GPDMAChannel[0].DMACCConfig & (1 << 17) );
    // disable I2S DMA
    LPC_GPDMAChannel[0].DMACCConfig &= ~1;
}


/*
 * Start I2S interface and codec with given sample rate.
 */
static void i2sStart( uint32_t sample_rate )
{
    fifiaudio.timerModulus = ((configCPU_CLOCK_HZ / 1000ul) * I2S_NUM_SAMPLES) / (4 * (fifiaudio.current_sample_rate/1000ul));
    LPC_TIM3->MR0 = fifiaudio.timerModulus - 1;
    LPC_TIM3->TC = 0; /* Reset counter */

    /* Start DMA */
    LPC_GPDMA->DMACIntTCClr = 1 << 0;
    LPC_GPDMA->DMACIntErrClr = 1 << 0;
    LPC_GPDMAChannel[0].DMACCSrcAddr = audioLli.source;
    LPC_GPDMAChannel[0].DMACCDestAddr = audioLli.destination;
    LPC_GPDMAChannel[0].DMACCLLI = audioLli.next;
    LPC_GPDMAChannel[0].DMACCControl = audioLli.control;
    LPC_GPDMAChannel[0].DMACCConfig =
            DMAC_CONFIG_WORD(
              DMAC_ID_I2S_0,
              DMAC_ID_DUMMY,
              DMAC_FLOW_P2M_DMAC,
              DMAC_ERROR_INT_OFF,
              DMAC_TERMINAL_COUNT_INT_ON
            );
    dmaSetCallback (0, fifiaudio_dmaCallbackFromISR);

	#if AUDIO_SAMPLES == 16
		// TX: 31 bits/half-period, muted, 16-bit data, master, reset
		LPC_I2S->I2SDAO = 	(1 << 15) |		// mute = 1
                            (31 << 6) |		// word select half period (-1), 31 = 64 Clocks
                            (1 << 4) |		// asynch reset of tx ch & i2s fifo
                            (1 << 0);		// wordwidth = 16 Bit

		// RX: 31 bits/half-period, 16-bit data, slave, reset
		LPC_I2S->I2SDAI = 	(31 << 6) |		// ws_halfperiod (-1) = 31 (64 Clocks)
                            (1 << 5) |		// ws_sel = slave
                            (1 << 4) | 		// asnych reset of rx ch & i2s fifo
                            (1 << 0);		// wordwidth = 16 Bit
	#endif
	#if AUDIO_SAMPLES == 32
		/* TX: 31 bits/half-period, muted, 32-bit data, master, reset */
		LPC_I2S->I2SDAO = (1 << 15) | (31 << 6) | (1 << 4) | (3 << 0);

		/* RX: 31 bits/half-period, 32-bit data, slave, reset */
		LPC_I2S->I2SDAI = (31 << 6) | (1 << 5) | (1 << 4) | (3 << 0);
	#endif

		/* Bit rate is 1/4 of MCLK */
		LPC_I2S->I2STXBITRATE = (4 - 1);
		LPC_I2S->I2SRXBITRATE = (4 - 1);

		/* TX: "Normal" mode, MCLK enabled */
		LPC_I2S->I2STXMODE = (1 << 3);	// TxMcEna

		/* RX. 4-wire mode */
		LPC_I2S->I2SRXMODE = 0x04;

		/* Set TX MCLK to 256*(sample rate) */
		if (sample_rate == 96000)
		{
			LPC_I2S->I2STXRATE = (68 << 8) | (69 << 0);   /* 96 kHz */
		}
		else
		{
			LPC_I2S->I2STXRATE = (68 << 8) | (138 << 0);   /* Default for 48 kHz */
		}

    /* Enable DMA1 for RX, DMA request threshold = 1 */
    LPC_I2S->I2SDMA1 = 	(1 << 0) |		// enable DMA for I2S RX
                        (4 << 8);		// trigger level for DMA rx req = 4

	LPC_I2S->I2SDAO &= ~(1 << 4);	// de-assert reset
	LPC_I2S->I2SDAI &= ~(1 << 4);	// de-assert reset


	//TODO Nicht hier initialisieren. DMA könnte schon laufen!
	LPC_USB->USBDMARClr = (1 << 7);
	LPC_USB->USBEpDMAEn = (1 << 7);


	usb_isoc_packets[0] = 	(1 << 16) |	// Frame #1
							20;			// Length = 33
	// Initialize USB DMA descriptor for isochronous transfer of audio data
	usb_dma_desc[0].next = &usb_dma_desc[1];
	usb_dma_desc[0].value1 = 	(1 << 2) |		// next DD valid = 1 (TODO ???)
								(1 << 4) |		// Isochronous endpoint
								(0x01 << 16);	// DMA Buffer length in packets
	usb_dma_desc[0].dma_buffer_start = audioBuffer;
	usb_dma_desc[0].value3 = 0;					// reset status etc to 0
	/*
	Only the Packet_length field of the isochronous packet size word is used. For each frame,
	an isochronous data packet of size specified by this field is transferred from the USB
	device to the host, and Isochronous_packet_size_memory_address is incremented by 4
	at the end of the packet transfer. If Packet_length is zero, an empty packet will be sent by
	the USB device.
	...
	an isochronous data packet of size specified by this field is transferred from the USB
	device to the host, and Isochronous_packet_size_memory_address is incremented by 4
	at the end of the packet transfer. If Packet_length is zero, an empty packet will be sent by
	the USB device.
	 */
	usb_dma_desc[0].isoc_packetsize_memory_address = &usb_isoc_packets[0];

    usb_isoc_packets[1] =   (1 << 16) |         // Frame #1
                            33;                 // Length = 33

    usb_dma_desc[1].next = &usb_dma_desc[0];
    usb_dma_desc[1].value1 = (1 << 2) |         // next DD valid = 1 (TODO ???)
                             (1 << 4) |         // Isochronous endpoint
                             (0x0001 << 16);    // DMA Buffer length in packets
    usb_dma_desc[1].dma_buffer_start = audioBuffer;
    usb_dma_desc[1].value3 = 0;
    usb_dma_desc[1].isoc_packetsize_memory_address = &usb_isoc_packets[1];

    usb_udca[2*3+1] = &usb_dma_desc[0];
}



/** Set volume control of UDA1361 (0/+6 dB switch). */
static void setUdaVolume (int16_t volume)
{
    if (volume == 0) {
        /* 0 dB volume in audio function */
        LPC_GPIO4->FIOSET = 0x10000000;
        LPC_GPIO4->FIODIR |= 0x10000000;    /* P4.28=UDA1361_PWON (high = 6 dB for UDA1361) */
    }
    else {
        /* -6 dB volume in audio function */
        LPC_GPIO4->FIODIR &= ~0x10000000;   /* P4.28=UDA1361_PWON (floating = 0 dB for UDA1361) */
    }
}



/**
  Handle class requests

  @param [in] pSetup
  @param [out] piLen
  @param [out] ppbData
 */
static bool HandleClassRequest (TSetupPacket *pSetup, int *piLen, void **ppData, portBASE_TYPE *taskWoken)
{
    (void)piLen;
    uint8_t *pbData = (uint8_t *)*ppData;
    int16_t *pwData = (int16_t *)*ppData;
    usb_Message message;
    uint8_t endpointNumber;
    uint8_t controlSelector;
    uint8_t unitId;


    if (REQTYPE_GET_RECIP(pSetup->bmRequestType) == REQTYPE_RECIP_INTERFACE) {
        controlSelector = (pSetup->wValue >> 8) & 0xFF;
        unitId = (pSetup->wIndex >> 8) & 0xFF;

        /* Volume control of codec (0/-6dB switch) */

        if (CONTROL_SELECTOR(USB_UNITID_IQ_FEATURE, USBAC_CS_FU_VOLUME_CONTROL, 2)) {

            switch (pSetup->bRequest) {
            case GET_MIN:   pwData[0] = -6 * dB;
                            break;
            case GET_MAX:   pwData[0] = 0 * dB;
                            break;
            case GET_CUR:   pwData[0] = fifiaudio.currentVolume;
                            break;
            case SET_CUR:   message.opcode = USBMSG_SET_VOLUME;
                            message.volume = pwData[0];
                            xQueueSendToBackFromISR(usb_queue, &message, taskWoken);
                            break;
            case GET_RES:   pwData[0] = 6 * dB;
                            break;
            case SET_RES:   /* Accept everything */
                            break;
            default:        return false;
            }
        }
        else {
            return false;
        }
    }

    /* Endpoint control: Set/Get sample rate */
    if (REQTYPE_GET_RECIP(pSetup->bmRequestType) == REQTYPE_RECIP_ENDPOINT) {
        endpointNumber = pSetup->wIndex & 0x8F;

        switch (pSetup->bRequest) {
        case SET_CUR:
            if ((pSetup->wLength == 3) &&
                ((pSetup->wValue >> 8) == USBAC_CS_EP_SAMPLING_FREQ_CONTROL)) {

                if (endpointNumber == USB_ENDPOINT_IQ) {
                    fifiaudio.current_sample_rate =
                        pbData[0] | (pbData[1] << 8) | (pbData[2] << 16);
                    message.opcode = USBMSG_I2S_CHANGE_SRATE;
                    xQueueSendToBackFromISR (usb_queue, &message, taskWoken);
                }
            }
            else {
                return false;
            }
            break;

        case GET_CUR:
            if ((pSetup->wLength == 3) &&
                ((pSetup->wValue >> 8) == USBAC_CS_EP_SAMPLING_FREQ_CONTROL)) {
                if (endpointNumber == USB_ENDPOINT_IQ) {
                    pbData[0] = fifiaudio.current_sample_rate & 0xFF;
                    pbData[1] = (fifiaudio.current_sample_rate >> 8) & 0xFF;
                    pbData[2] = (fifiaudio.current_sample_rate >> 16) & 0xFF;
                }
            }
            else {
                return false;
            }
            break;
        }
    }

    return true;
}



/** Return current volume. */
int16_t usbclass_getVolume (void)
{
    return fifiaudio.currentVolume;
}



static void USBFrameHandler (unsigned short wFrame, portBASE_TYPE *taskWoken)
{
    usb_Message message;

    message.opcode = USBMSG_FRAME;
    message.frame = wFrame;
    xQueueSendToBackFromISR(usb_queue, &message, taskWoken);
}




static void usbclass_frameHandler (uint16_t frame __attribute__((unused)))
{
    int32_t temp;
    int32_t currentPos;
    int32_t delta;
    int32_t i;


    /* Do nothing if channel isn't active */
    if (!fifiaudio.i2s_active) {
        return;
    }

    /* Take timer snapshot */
    temp = (int32_t)LPC_TIM3->TC;

    /* Are we in sync?
     * Sync is achieved if the read index is half-way from the current write position,
     * plus/minus some samples.
     */

    /* The next calculation works for both cases: current > start and current < start.
     * The addition of the modulus guarantees a positive result.
     * Note: It is possible that the DMA interrupt fires at this very moment, and updates the
     *       lastDmaTimestamp. This is no problem, since the timer modulus is chosen to be
     *       exactly one DMA period, and therefore it is expected that the lastDmaTimestamp
     *       will not change (significantly).
     */
    currentPos = fifiaudio.timerModulus + temp - fifiaudio.lastDmaTimestamp;
    currentPos %= fifiaudio.timerModulus;

    /* Convert to #samples: pos /= (CCLK/4) / samplerate
     * Move to diameter: pos = (pos + NUM_SAMPLES/2) % NUM_SAMPLES
     * The calculation below does it all in one without rounding and overflow problems.
     */
    currentPos = (2 * currentPos * (fifiaudio.current_sample_rate/1000ul) +
                  I2S_NUM_SAMPLES * (configCPU_CLOCK_HZ / 4000ul))  /
                  (configCPU_CLOCK_HZ / 2000ul);
    currentPos %= I2S_NUM_SAMPLES;

    /* Calculate difference in the range -512...+511. Nominal value is 0. */
    temp = (((3*I2S_NUM_SAMPLES)/2 + fifiaudio.readIndex - currentPos) % I2S_NUM_SAMPLES);
    temp -= I2S_NUM_SAMPLES / 2;

    /* Calculate necessary adjustment. */
    delta = 0;
    if ((temp < -50) || (temp > +50)) {
        /* Too far off. Force sync, and accept losing samples. */
        fifiaudio.readIndex = currentPos;
    }
    else if (temp < -5) {
        /* Internal clock too fast. Long packet (+1 sample) */
        delta = +1;
    }
    else if (temp > +5) {
        /* Internal clock too slow. Short packet (-1 sample) */
        delta = -1;
    }

    /* Copy from start of buffer to beyond the buffer end if necessary. */
    temp = fifiaudio.readIndex + (fifiaudio.current_sample_rate/1000ul) + delta;
    if (temp > I2S_NUM_SAMPLES) {
        while (temp - I2S_NUM_SAMPLES) {
            temp--;
            audioBuffer[temp] = audioBuffer[temp - I2S_NUM_SAMPLES];
        }
    }

    i = (usb_dma_desc[0].value3 & 1) ? 0 : 1;

    usb_isoc_packets[i] = (1 << 16) | (4 * (fifiaudio.current_sample_rate/1000ul + delta));

    usb_dma_desc[i].value1 = (1 << 2) | (1 << 4) | (0x0001 << 16);
    usb_dma_desc[i].dma_buffer_start = &audioBuffer[fifiaudio.readIndex];
    usb_dma_desc[i].isoc_packetsize_memory_address = &usb_isoc_packets[i];
    usb_dma_desc[i].value3 = 0;

    /* Underrun? */
    if (LPC_USB->USBNDDRIntSt & (1 << 7)) {
        /* Clear it */
        LPC_USB->USBNDDRIntClr = (1 << 7);
    }

    /* Advance read pointer */
    fifiaudio.readIndex += ((fifiaudio.current_sample_rate/1000ul) + delta);
    fifiaudio.readIndex %= I2S_NUM_SAMPLES;
}



/** Inform class about setup request (only vendor for now...). */
void usbclass_setupRequestFromISR (TSetupPacket *pSetup, void **ppData, portBASE_TYPE *taskWoken)
{
    usb_Message message;

    message.opcode = USBMSG_SETUP_REQUEST;
    message.setup = *pSetup;
    message.ppData = ppData;
    xQueueSendToBackFromISR(usb_queue, &message, taskWoken);
}



/** Inform task about frame interrupt. */
void usbclass_frameFromISR (uint16_t frame, portBASE_TYPE *taskWoken)
{
    usb_Message message;

    message.opcode = USBMSG_FRAME;
    message.frame = frame;
    xQueueSendToBackFromISR(usb_queue, &message, taskWoken);
}


void usb_task (void *pvParameters __attribute__((unused)))
{
    usb_Message message;
    int ignoreFrame = 0;


    /* Timer3 is used as a high-resolution timing reference to help detect clock drift. */
    LPC_SC->PCONP |= (1ul << 23);
    LPC_TIM3->TCR = 2;                      /* Stop and reset */
    LPC_TIM3->PR = 0;                       /* Maximum resolution */
    LPC_TIM3->MR0 = 1;
    LPC_TIM3->MCR = 2;                      /* Reset by MR0 */
    LPC_TIM3->TCR = 1;                      /* Start */

    fifiaudio.current_sample_rate = 48000ul;
    fifiaudio.timerModulus = 532000ul;

    /* Task communication */
    usb_queue = xQueueCreate (USBQUEUE_LENGTH, sizeof(usb_Message));


    // initialize stack
    // set StdRequestHandler
    usbd_base_init();

LPC_USB->USBUDCAH = (uint32_t)&usb_udca;

    // register class request handler
    USBRegisterRequestHandler(REQTYPE_TYPE_CLASS, HandleClassRequest, abClassReqData);

    // I2S interface
    dmaInit();
    //TODO: Fehlerbehandlung wenn DMA nicht gestartet werden kann (LED?!)
    while( dmaResume() == DMAC_BUSY );  // DMA has to be started

    // register frame handler
    USBHwRegisterFrameHandler(USBFrameHandler);

    // enable bulk-in interrupts on NAKs
    USBHwNakIntEnable(INACK_BI);

softrock_init();

    usbd_start();

    // connect to bus
    USBHwConnect(true);

    while (1) {
        if (xQueueReceive( usb_queue, &message, portMAX_DELAY ) == pdTRUE) {
            switch (message.opcode) {
                case USBMSG_I2S_CHANGE_SRATE:
                    if (fifiaudio.i2s_active) {
                        i2sStart(fifiaudio.current_sample_rate);
                    }
                break;

                case USBMSG_I2S_START:
                    USBHwEPConfig(AUDIO_IN_EP, USB_ISOC_SIZE_IQ);

                    i2sStart(fifiaudio.current_sample_rate);
                    fifiaudio.i2s_active = 1;
                break;

                case USBMSG_I2S_STOP:
                    USBHwEPConfig(AUDIO_IN_EP, 0);

                    i2sStop();
                    fifiaudio.i2s_active = 0;
                break;

                case USBMSG_SETUP_REQUEST:
                    softrock_handleRequest(&message.setup, message.ppData);
                break;

                case USBMSG_FRAME:
                    if (ignoreFrame) {
                        ignoreFrame--;
                        if (ignoreFrame == 0) {
                            /* TODO: Notnagel... ohne das gibt es Klicks nach einem paramsWrite() */
                            message.opcode = USBMSG_I2S_START;
                            xQueueSendToBack(usb_queue, &message, 0);
                        }
                    } else {
                        usbclass_frameHandler(message.frame);
                    }
                break;

                case USBMSG_I2S_HOLD:
                    ignoreFrame = 200;
                break;

                case USBMSG_SET_VOLUME:
                    fifiaudio.currentVolume = message.volume;
//                    if (fifiaudio.currentVolume) {
//                        fifiaudio.currentVolume = 0x0600;   /* Only one non-zero value allowed */
//                    }
                    setUdaVolume(fifiaudio.currentVolume);
                break;

                case USBMSG_SET_PREAMP:
                    switch (message.mode) {
                        case 0:
                            fifiaudio.currentVolume = -6 * dB;
                            setUdaVolume(fifiaudio.currentVolume);
                        break;
                        case 1:
                            fifiaudio.currentVolume = 0;
                            setUdaVolume(fifiaudio.currentVolume);
                        break;
                    }
                break;
            }
        }
    }
}

